// -*- Mode: vala; indent-tabs-mode: nil; tab-width: 4 -*-
/***
    BEGIN LICENSE

    Copyright (C) 2011-2013 Wingpanel Developers
    This program is free software: you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License version 3, as published
    by the Free Software Foundation.

    This program is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranties of
    MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
    PURPOSE.  See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program.  If not, see <http://www.gnu.org/licenses/>

    END LICENSE
***/

using Gdk;

namespace SuperWingpanel.Widgets
{
    public class PanelWindow : Gtk.Window 
    {
        public PanelController controller { private get; construct; }

        private Gtk.Box container;
        private Gtk.Box left_wrapper;
        private Gtk.Box right_wrapper;
        private Gtk.SizeGroup size_group;

        private IndicatorMenubar indicator_menubar;
        private MenuBar clock_menubar;
        private MenuBar apps_menubar;

        private IndicatorLoader indicator_loader;


        public PanelWindow (WingpanelApp app, PanelController controller, IndicatorLoader indicator_loader)
        {
            GLib.Object (controller: controller);

            decorated = false;
            resizable = false;
            skip_taskbar_hint = true;
            app_paintable = true;
            set_visual (get_screen ().get_rgba_visual ());
            set_type_hint (Gdk.WindowTypeHint.DOCK);

            set_application (app as Gtk.Application);

            this.indicator_loader = indicator_loader;

            container = new Gtk.Box (Gtk.Orientation.HORIZONTAL,  0);
            left_wrapper = new Gtk.Box (Gtk.Orientation.HORIZONTAL, 0);
            right_wrapper = new Gtk.Box (Gtk.Orientation.HORIZONTAL, 0);
            container.set_homogeneous (false);
            left_wrapper.set_homogeneous (false);
            right_wrapper.set_homogeneous (false);

            add (container);

            var style_context = get_style_context ();
            style_context.add_class (StyleClass.PANEL);
            style_context.add_class (Gtk.STYLE_CLASS_MENUBAR);

            add_default_widgets (controller.settings);

            load_indicators ();

            controller.settings.notify["enable-slim-mode"].connect (slim_mode_changed);

            // Watch for mouse
            add_events (EventMask.ENTER_NOTIFY_MASK |
                        EventMask.LEAVE_NOTIFY_MASK |
                        EventMask.POINTER_MOTION_MASK);

            destroy.connect (Gtk.main_quit);
        }

        ~PanelWindow ()
        {
            controller.settings.notify["enabled-slim-mode"].disconnect (slim_mode_changed);
        }


        public Gtk.StyleContext get_draw_style_context ()
        {
            return indicator_menubar.get_style_context ();
        }


        private void add_default_widgets (Services.Settings settings)
        {
            // Add Apps button
            apps_menubar = new MenuBar ();
            var apps_button = new Widgets.AppsButton (settings);
            apps_menubar.append (apps_button);

            left_wrapper.pack_start (apps_menubar, false, true, controller.renderer.SLIM_PANEL_EDGE_PADDING);

            container.pack_start (left_wrapper);

            clock_menubar = new MenuBar ();
            container.pack_start (clock_menubar, false, false, 0);

            // Menubar for storing indicators
            indicator_menubar = new IndicatorMenubar ();

            right_wrapper.pack_end (indicator_menubar, false, false, controller.renderer.SLIM_PANEL_EDGE_PADDING);
            container.pack_end (right_wrapper);

            size_group = new Gtk.SizeGroup (Gtk.SizeGroupMode.HORIZONTAL);
            set_clock_centered (!controller.settings.enable_slim_mode);
            
        }

        public void set_clock_centered (bool enable)
        {
            if (enable)
            {
                size_group.add_widget (left_wrapper);
                size_group.add_widget (right_wrapper);
            }
            else 
            {
                size_group.remove_widget (left_wrapper);
                size_group.remove_widget (right_wrapper);       
            }
        }

        private void load_indicators () {
            var indicators = indicator_loader.get_indicators ();

            foreach (var indicator in indicators)
                load_indicator (indicator);
        }

        private void load_indicator (IndicatorIface indicator)
        {
            var entries = indicator.get_entries ();

            foreach (var entry in entries)
                create_entry (entry);

            indicator.entry_added.connect (create_entry);
            indicator.entry_removed.connect (delete_entry);
        }

        private void create_entry (IndicatorWidget entry)
        {
            if (entry.get_indicator ().get_name () == "libsynapse.so")
            {
                apps_menubar.append (entry);
            }
            else if (entry.get_indicator ().get_name () == "libdatetime.so")
            {
                clock_menubar.prepend (entry);
            }
            else
            {
                indicator_menubar.insert_sorted (entry);
            }
        }

        private void delete_entry (IndicatorWidget entry)
        {
            var parent = entry.parent;
            parent.remove (entry);
        }




        public void slim_mode_changed ()
        {
            update_size_and_position ();
        }        


        public void update_size_and_position ()
        {
            unowned PositionManager position_manager = controller.position_manager;
            position_manager.update_size_and_position ();

            update_clock_alignment ();

            // TODO: animate these
            move (position_manager.X, position_manager.Y);
            set_size_request (position_manager.W, position_manager.H);
        }


        public void update_clock_alignment ()
        {
            if (controller.settings.enable_slim_mode)
            {
                set_clock_centered (false);
            }
            else if (controller.settings.hide_mode == HideType.INTELLISLIM && controller.renderer.hide_progress >= 0.5)
            {
                set_clock_centered (false);
            }
            else
            {
                set_clock_centered  (true);
            }
        }

        public override bool draw (Cairo.Context cr) 
        {
            controller.renderer.draw_panel (cr);
            return true;
        }
        
        public bool menu_is_visible ()
        {
            return false;
        }
        
    }

}